#include "cntl2.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

/*************Local Function Declarations***************/
#define BLOCK_STATUS_FREE		0
#define BLOCK_STATUS_GETTING	1
#define BLOCK_STATUS_DONE		2

int hm_read_config(char* config_file);

/*******************************************************
* Starup everyting
* fname is the name of the config file
*******************************************************/
int hm_init(char* fname){
	struct sockaddr_in	sock_addr;
	int					i;

	/*init structures*/
	bzero(Jobs,    sizeof(JobRecord)   *MAX_JOBS);
	bzero(Packets, sizeof(PacketRecord)*MAX_PACKETS);
	bzero(Targets, sizeof(TargetRecord)*MAX_TARGETS);

	/*open the socket*/
	GlobalFD=socket(AF_INET, SOCK_DGRAM, 0);
	sock_addr.sin_family=AF_INET;
	sock_addr.sin_port=0;
	sock_addr.sin_addr.s_addr=INADDR_ANY;
	bzero(&sock_addr.sin_zero, 8);
	bind(GlobalFD, (struct sockaddr*)&sock_addr,sizeof(struct sockaddr));
	fcntl(GlobalFD, F_SETFL, O_NONBLOCK);
	
#ifdef DEBUG
	printf("Socket is %i\n",GlobalFD);
#endif	

#ifdef KEEP_UNTIL_I_ERASE	
	/*init the keys*/
	for (i=0;i<MAX_TARGETS;i++){
		makeKey(&Targets[i].Key.ki, DIR_ENCRYPT, KeySize, NULL);
		cipherInit(&Targets[i].Key.ci, MODE_ECB, NULL);
	}	
#endif
	CryptoInit();

	/*read in the config file*/
	if (!hm_read_config(fname)) return FALSE;
	
	return TRUE;
}

/******************************************************
* Shut everything down
******************************************************/
void hm_close(){
	close(GlobalFD);
}

/***************************************************
* Find the encryption key that goes with this IP
***************************************************/
TargetRecord* hm_get_target_from_ip(struct in_addr IP){
	int i;
	
	for (i=0;i<MAX_TARGETS;i++){
		if (memcmp(&Targets[i].IP, &IP, sizeof(struct in_addr))==0){
			return &Targets[i];
		} 
	}
	printf("Couldn't resolve ip/key\n");
	return NULL;
}

/*****************************************************
* Resend the contents of a PacketRecord if it needs rent
* Expire it if needed
*****************************************************/
int hm_resend_packet(PacketRecord* p){
	struct sockaddr_in			sa;
	struct timezone				tz;
	struct timeval				tv;

	gettimeofday(&tv, &tz);
	
	if (p->LastTime==0 || (tv.tv_sec-p->LastTime)>RETRY_TIME){ 
		sa.sin_family=AF_INET;
		sa.sin_port=htons(5757);
		sa.sin_addr=p->Job->Target->IP;
		bzero(&sa.sin_zero, 8);
	
		if (p->RetriesLeft==0){
			/*expire the packet*/
			p->Callback(p, FALSE, NULL, 0, p->User);
			return FALSE;
		}
	
		p->RetriesLeft--;	
		p->LastTime=tv.tv_sec;
		
		sendto(GlobalFD, p->RawPacket, p->PacketSize, 0, (struct sockaddr*)&sa, sizeof(struct sockaddr));
		return TRUE;
	}
	
	return FALSE;	
}

/*****************************************************
* Resend all the packets if needed
*****************************************************/
int hm_resend_all(){
	int i;
	int	did_resend;
	
	did_resend=FALSE;
	for (i=0;i<MAX_PACKETS;i++){
		if (Packets[i].InUse)
		if (hm_resend_packet(&Packets[i])) did_resend=TRUE;
	}
	
	return did_resend;
}


/*****************************************************
* This gets called when a stat job gets a packet
*****************************************************/
int hm_stat_packet(PacketRecord* Packet, int Success, char* Data, int DataLen, void* User){

	Packet->Job->BlockingResult=Success;	
	if (Success){
		/*copy the stat packet for return*/
		memcpy( ((JobStat*)(Packet->Job->JobVars))->Stat, Data, DataLen);
	}
	
	/*call the Job callback*/
	if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
	Packet->Job->InUse=FALSE;
}

/*****************************************************
* This gets called when a ping job gets a packet
*****************************************************/
int hm_ping_packet(PacketRecord* Packet, int Success, char* Data, int DataLen, void* User){

	Packet->Job->BlockingResult=Success;	

	/*call the Job callback*/
	if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
}

/*****************************************************
* This gets called when a get buff job gets a packet
* TODO: Check Flags and stuff
*****************************************************/
int hm_get_buff_packet(PacketRecord* Packet, int Success, char* Data, int DataLen, void* User){
	ControlPacketBuffReply*		p;
	JobGetBuff*					j;
	ControlPacketBuffRequest	pb;
	int							count;
	int							next_chunk;
	int							total_chunk;
	int							more_chunks;
	
	if (Success){
		if (!Data) return FALSE;
	
		j=(JobGetBuff*)Packet->Job->JobVars;
		p=(ControlPacketBuffReply*)Data;
		total_chunk=(j->TargetBuffLen/MAX_BUFFER_SEGMENT)+1;
		
		/*copy the data into the buffer*/
#ifdef DEBUG
		printf("j             %p\n",j);
		printf("p             %p\n",p);
		printf("j->TargetBuff %p\n",j->TargetBuff);
		printf("p->Offset     %u\n",p->Offset);
		printf("p->Data       %p\n",p->Data);
		printf("p->Len        %u\n",p->Len);
#endif		
		memcpy(j->TargetBuff+p->Offset, p->Data, p->Len);
		
		/*mark that chunk as done*/
		//printf("marking chunk %i/%i as done\n",p->Offset/MAX_BUFFER_SEGMENT, total_chunk);
		j->BlockStatus[p->Offset/MAX_BUFFER_SEGMENT]=BLOCK_STATUS_DONE;
		
		/*find the next chunk to send*/
		more_chunks=FALSE; 
		for (next_chunk=0;next_chunk<total_chunk;next_chunk++){
			if (j->BlockStatus[next_chunk]==BLOCK_STATUS_GETTING) more_chunks=TRUE;
			if (j->BlockStatus[next_chunk]==BLOCK_STATUS_FREE) break;
		}  
		 
		if (next_chunk==total_chunk){
			Packet->InUse=FALSE;
			if (!more_chunks){
				/*job is all done*/
				Packet->Job->BlockingResult=Success;	
				/*call the Job callback*/
				if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
			}
			return TRUE;
		} 
		 
		 j->BlockStatus[next_chunk]=BLOCK_STATUS_GETTING;
		
		/*Send off the next packet*/
		/*build a packet*/
		Packet->InUse=TRUE;
		bzero(Packet->RawPacket, MAX_PACKET_SIZE);
		Packet->PacketSize=sizeof(ControlPacketBuffRequest);
		Packet->PacketID=random()*65535*65535;
		Packet->RetriesLeft=NUM_RETRIES+1;
		Packet->LastTime=0;
		Packet->Job=Packet->Job;
		Packet->Callback=hm_get_buff_packet;
		
		pb.MagicToken1=CONTROL_MAGIC_TOKEN;
		pb.MagicToken2=CONTROL_MAGIC_TOKEN;
		pb.PacketType=CONTROL_PACKET_BUFF_REQUEST;
		pb.PacketID=Packet->PacketID;
		pb.BuffID=((JobGetBuff*)Packet->Job->JobVars)->BuffID;
		pb.Offset=next_chunk*MAX_BUFFER_SEGMENT;
		pb.Len=MAX_BUFFER_SEGMENT;
		pb.Flags=0;

#ifdef HOLD_UNTIL_I_ERASE		
		/*encrypt the packet*/
		*(unsigned int*)Packet->RawPacket=CONTROL_MAGIC_TOKEN;
		count=blockEncrypt(&Packet->Job->Target->Key.ci, &Packet->Job->Target->Key.ki, (char*)&pb+4, (sizeof(ControlPacketBuffRequest)-4)*8, Packet->RawPacket+4);
		if (count!=(sizeof(ControlPacketBuffRequest)-4)*8){
			printf("Failed to encrypt returned %li\n",count);
			return -1;
		}
#endif				
		((unsigned int*)Packet->RawPacket)[0]=CONTROL_MAGIC_TOKEN;
		Packet->Job->Target->Crypto->Encrypt((struct crypto_record*)Packet->Job->Target->Crypto, (char*)&pb+4, Packet->RawPacket+4, sizeof(ControlPacketBuffRequest)-4);

		


		/*send it off*/
		hm_resend_packet(Packet);
	}else{
		Packet->Job->BlockingResult=Success;	
		/*call the Job callback*/
		if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
	}
}

/*****************************************************
* This gets called when a get buff job gets a packet
* TODO: Check Flags and stuff
*****************************************************/
int hm_set_buff_packet(PacketRecord* Packet, int Success, char* Data, int DataLen, void* User){
	ControlPacketSetBufferReply*		p;
	JobSetBuff*							j;
	ControlPacketSetBuffer				pb;
	int									count;
	int									next_chunk;
	int									total_chunk;
	int									more_chunks;
		
	if (Success){
		j=(JobSetBuff*)Packet->Job->JobVars;
		p=(ControlPacketSetBufferReply*)Data;
		total_chunk=(j->TargetBuffLen/MAX_BUFFER_SEGMENT)+1;
		
		/*mark that chunk as done*/
		//printf("marking chunk %i/%i as done\n",p->Offset/MAX_BUFFER_SEGMENT, total_chunk);
		j->BlockStatus[p->Offset/MAX_BUFFER_SEGMENT]=BLOCK_STATUS_DONE;
		
		/*find the next chunk to send*/
		more_chunks=FALSE; 
		for (next_chunk=0;next_chunk<total_chunk;next_chunk++){
			if (j->BlockStatus[next_chunk]==BLOCK_STATUS_GETTING) more_chunks=TRUE;
			if (j->BlockStatus[next_chunk]==BLOCK_STATUS_FREE) break;
		}  
		 
		if (next_chunk==total_chunk){
			Packet->InUse=FALSE;
			if (!more_chunks){
				/*job is all done*/
				Packet->Job->BlockingResult=Success;	
				/*call the Job callback*/
				if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
			}
			return TRUE;
		} 
		 
		 j->BlockStatus[next_chunk]=BLOCK_STATUS_GETTING;
		
		/*Send off the next packet*/
		/*build a packet*/
		Packet->InUse=TRUE;
		bzero(Packet->RawPacket, MAX_PACKET_SIZE);
		Packet->PacketSize=sizeof(ControlPacketSetBuffer);
		Packet->PacketID=random()*65535*65535;
		Packet->RetriesLeft=NUM_RETRIES+1;
		Packet->LastTime=0;
		Packet->Job=Packet->Job;
		Packet->Callback=hm_set_buff_packet;
		
		pb.MagicToken1=CONTROL_MAGIC_TOKEN;
		pb.MagicToken2=CONTROL_MAGIC_TOKEN;
		pb.PacketType=CONTROL_PACKET_SET_BUFF;
		pb.PacketID=Packet->PacketID;
		pb.BuffID=((JobSetBuff*)Packet->Job->JobVars)->BuffID;
		pb.Offset=next_chunk*MAX_BUFFER_SEGMENT;
		if ((pb.Offset+MAX_BUFFER_SEGMENT)>j->TargetBuffLen){
			pb.Len=j->TargetBuffLen % MAX_BUFFER_SEGMENT;
		}else{
		 	pb.Len=MAX_BUFFER_SEGMENT;
		}	
		pb.Flags=0;
		memcpy(pb.Data, j->TargetBuff + pb.Offset, pb.Len); 
		
		((unsigned int*)Packet->RawPacket)[0]=CONTROL_MAGIC_TOKEN;
		Packet->Job->Target->Crypto->Encrypt((struct crypto_record*)Packet->Job->Target->Crypto, (char*)&pb+4, Packet->RawPacket+4, sizeof(ControlPacketSetBuffer)-4);
		

		/*send it off*/
		hm_resend_packet(Packet);
	}else{
		Packet->Job->BlockingResult=Success;	
		/*call the Job callback*/
		if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
	}
}

/*****************************************************
* This gets called when a command job gets a packet
*****************************************************/
int hm_command_packet(PacketRecord* Packet, int Success, char* Data, int DataLen, void* User){

	Packet->Job->BlockingResult=Success;	
	if (Success){
		/*copy the stat packet for return*/
		memcpy( ((JobCommand*)(Packet->Job->JobVars))->Reply, Data, DataLen);
	}
	
	/*call the Job callback*/
	if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
	Packet->Job->InUse=FALSE;
}


/*****************************************************
* This gets called when a sync gets a packet
* TODO: Check Flags and stuff
*****************************************************/
int hm_file_sync_packet(PacketRecord* Packet, int Success, char* Data, int DataLen, void* User){
	ControlPacketBuffReply*		p;
	JobFileSync*				j;
	ControlPacketBuffRequest	pb;
	int							count;
	int							next_chunk;
	int							total_chunk;
	int							more_chunks;
	
	if (Success){
		j=(JobFileSync*)Packet->Job->JobVars;
		p=(ControlPacketBuffReply*)Data;
		total_chunk=(j->BuffLen/MAX_BUFFER_SEGMENT)+1;
		
		/*write the data out to the file*/
		fseek(j->fp, p->Offset, SEEK_SET);
		fwrite(p->Data, p->Len, 1, j->fp);
		
		/*mark that chunk as done*/
#ifdef DEBUG		
		printf("marking chunk %i/%i as done\n",p->Offset/MAX_BUFFER_SEGMENT, total_chunk);
#endif		
		j->BlockStatus[p->Offset/MAX_BUFFER_SEGMENT]=BLOCK_STATUS_DONE;
		
		/*find the next chunk to send*/
		more_chunks=FALSE; 
		for (next_chunk=0;next_chunk<total_chunk;next_chunk++){
			if (j->BlockStatus[next_chunk]==BLOCK_STATUS_GETTING){
#ifdef DEBUG			
				printf("Block %i is still getting\n",next_chunk);
#endif				
				more_chunks=TRUE;
			}	
			if (j->BlockStatus[next_chunk]==BLOCK_STATUS_FREE) break;
		}  
		
#ifdef DEBUG		
		printf("next_chunk is %i total_chunk is %i\n",next_chunk, total_chunk);
#endif		
		 
		if (next_chunk==total_chunk){
			Packet->InUse=FALSE;
			if (!more_chunks){
				/*job is all done*/
				Packet->Job->BlockingResult=Success;	
				/*call the Job callback*/
				fclose(j->fp);
#ifdef DEBUG				
				printf("About to call callback\n");
#endif				
				if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
			}else{
#ifdef DEBUG			
				printf("There are more chunks\n");
#endif				
			}
			return TRUE;
		} 
		 
#ifdef DEBUG		 
		 printf("Getting next_chunk=%i\n",next_chunk);
#endif		 
		 j->BlockStatus[next_chunk]=BLOCK_STATUS_GETTING;
		
		/*Send off the next packet*/
		/*build a packet*/
		Packet->InUse=TRUE;
		bzero(Packet->RawPacket, MAX_PACKET_SIZE);
		Packet->PacketSize=sizeof(ControlPacketBuffRequest);
		Packet->PacketID=random()*65535*65535;
		Packet->RetriesLeft=NUM_RETRIES+1;
		Packet->LastTime=0;
		Packet->Job=Packet->Job;
		Packet->Callback=hm_file_sync_packet;
		
		pb.MagicToken1=CONTROL_MAGIC_TOKEN;
		pb.MagicToken2=CONTROL_MAGIC_TOKEN;
		pb.PacketType=CONTROL_PACKET_BUFF_REQUEST;
		pb.PacketID=Packet->PacketID;
		pb.BuffID=((JobGetBuff*)Packet->Job->JobVars)->BuffID;
		pb.Offset=next_chunk*MAX_BUFFER_SEGMENT;
		pb.Len=MAX_BUFFER_SEGMENT;
		pb.Flags=0;
		
		((unsigned int*)Packet->RawPacket)[0]=CONTROL_MAGIC_TOKEN;
		Packet->Job->Target->Crypto->Encrypt((struct crypto_record*)Packet->Job->Target->Crypto, (char*)&pb+4, Packet->RawPacket+4, sizeof(ControlPacketBuffRequest)-4);

		/*send it off*/
		hm_resend_packet(Packet);
	}else{
		Packet->Job->BlockingResult=Success;	
		/*call the Job callback*/
		//j=(JobFileSync*)Packet->Job->JobVars;
		//fclose(j->fp);
		if (Packet->Job->Callback) Packet->Job->Callback(Success, Packet->Job->JobVars);
	}
}

/*****************************************************
* Start a stat job running
*****************************************************/
int hm_job_stat_start(JobRecord* job){
	int							i;
	ControlPacketStatRequest	p;
	int							count;
		
	/*find a free packet*/
	for (i=0;i<MAX_PACKETS;i++) if (!Packets[i].InUse) break;
	if (i==MAX_PACKETS) return -1;
	
	/*build a packet*/
	Packets[i].InUse=TRUE;
	bzero(Packets[i].RawPacket, MAX_PACKET_SIZE);
	Packets[i].PacketSize=sizeof(ControlPacketStatRequest);
	Packets[i].PacketID=random()*65535*65535;
	Packets[i].RetriesLeft=NUM_RETRIES+1;
	Packets[i].LastTime=0;
	Packets[i].Job=job;
	Packets[i].Callback=hm_stat_packet;
	
	p.MagicToken1=CONTROL_MAGIC_TOKEN;
	p.MagicToken2=CONTROL_MAGIC_TOKEN;
	p.PacketType=CONTROL_PACKET_STAT_REQUEST;
	p.BuffID=((JobStat*)job->JobVars)->BuffID;
	p.PacketID=Packets[i].PacketID;

	((unsigned int*)Packets[i].RawPacket)[0]=CONTROL_MAGIC_TOKEN;
	job->Target->Crypto->Encrypt((struct crypto_record*)job->Target->Crypto, (char*)&p+4, Packets[i].RawPacket+4, sizeof(ControlPacketStatRequest)-4);
	
	/*send it off*/
	return hm_resend_packet(&Packets[i]);
}

/*****************************************************
* Start a ping job running
*****************************************************/
int hm_job_ping_start(JobRecord* job){
	int							i;
	ControlPacketPing			p;
	int							count;
		
	/*find a free packet*/
	for (i=0;i<MAX_PACKETS;i++) if (!Packets[i].InUse) break;
	if (i==MAX_PACKETS) return -1;
	
	/*build a packet*/
	Packets[i].InUse=TRUE;
	bzero(Packets[i].RawPacket, MAX_PACKET_SIZE);
	Packets[i].PacketSize=sizeof(ControlPacketPing);
	Packets[i].PacketID=random()*65535*65535;
	Packets[i].RetriesLeft=NUM_RETRIES+1;
	Packets[i].LastTime=0;
	Packets[i].Job=job;
	Packets[i].Callback=hm_ping_packet;
	
	p.MagicToken1=CONTROL_MAGIC_TOKEN;
	p.MagicToken2=CONTROL_MAGIC_TOKEN;
	p.PacketType=CONTROL_PACKET_PING;
	snprintf(p.PingMessage, 244, ((JobPing*)job->JobVars)->Message);
	p.PacketID=Packets[i].PacketID;
	
	
	((unsigned int*)Packets[i].RawPacket)[0]=CONTROL_MAGIC_TOKEN;
	job->Target->Crypto->Encrypt((struct crypto_record*)job->Target->Crypto, (char*)&p+4, Packets[i].RawPacket+4, sizeof(ControlPacketPing)-4);

	/*send it off*/
	return hm_resend_packet(&Packets[i]);
}


/*****************************************************
* Start a get buffer job running
*****************************************************/
int hm_job_get_buff_start(JobRecord* job){
	int							i,j;
	ControlPacketBuffRequest	p;
	int							count;
	int							num_packets;
		
	/*send off a bunch of packets*/
	num_packets=(((JobGetBuff*)job->JobVars)->TargetBuffLen/MAX_BUFFER_SEGMENT) +1;
	if (num_packets>MAX_PACKET_PER_JOB) num_packets=MAX_PACKET_PER_JOB;
		
	/*find a free packet*/
	for (j=0;j<num_packets;j++){
		for (i=0;i<MAX_PACKETS;i++) if (!Packets[i].InUse) break;
		if (i==MAX_PACKETS) return -1;
		
		/*mark the packet as getting*/
		((JobGetBuff*)job->JobVars)->BlockStatus[j]=BLOCK_STATUS_GETTING;
		
		/*build a packet*/
		Packets[i].InUse=TRUE;
		bzero(Packets[i].RawPacket, MAX_PACKET_SIZE);
		Packets[i].PacketSize=sizeof(ControlPacketBuffRequest);
		Packets[i].PacketID=random()*65535*65535;
		Packets[i].RetriesLeft=NUM_RETRIES+1;
		Packets[i].LastTime=0;
		Packets[i].Job=job;
		Packets[i].Callback=hm_get_buff_packet;
		
		p.MagicToken1=CONTROL_MAGIC_TOKEN;
		p.MagicToken2=CONTROL_MAGIC_TOKEN;
		p.PacketType=CONTROL_PACKET_BUFF_REQUEST;
		p.PacketID=Packets[i].PacketID;
		p.BuffID=((JobGetBuff*)job->JobVars)->BuffID;
		p.Offset=j*MAX_BUFFER_SEGMENT;
		p.Len=MAX_BUFFER_SEGMENT;
		p.Flags=0;

		((unsigned int*)Packets[i].RawPacket)[0]=CONTROL_MAGIC_TOKEN;
		job->Target->Crypto->Encrypt((struct crypto_record*)job->Target->Crypto, (char*)&p+4, Packets[i].RawPacket+4, sizeof(ControlPacketBuffRequest)-4);

		/*send it off*/
		hm_resend_packet(&Packets[i]);
	}	
}

/*****************************************************
* Start a set buffer job running
*****************************************************/
int hm_job_set_buff_start(JobRecord* job){
	int							i,j;
	ControlPacketSetBuffer		p;
	int							count;
	int							num_packets;
		
	/*send off a bunch of packets*/
	num_packets=(((JobSetBuff*)job->JobVars)->TargetBuffLen/MAX_BUFFER_SEGMENT) +1;
	if (num_packets>MAX_PACKET_PER_JOB) num_packets=MAX_PACKET_PER_JOB;
		
	/*find a free packet*/
	for (j=0;j<num_packets;j++){
		for (i=0;i<MAX_PACKETS;i++) if (!Packets[i].InUse) break;
		if (i==MAX_PACKETS) return -1;
		
		/*mark the packet as getting*/
		((JobSetBuff*)job->JobVars)->BlockStatus[j]=BLOCK_STATUS_GETTING;
		
		/*build a packet*/
		Packets[i].InUse=TRUE;
		bzero(Packets[i].RawPacket, MAX_PACKET_SIZE);
		Packets[i].PacketSize=sizeof(ControlPacketSetBuffer);
		Packets[i].PacketID=random()*65535*65535;
		Packets[i].RetriesLeft=NUM_RETRIES+1;
		Packets[i].LastTime=0;
		Packets[i].Job=job;
		Packets[i].Callback=hm_set_buff_packet;
		
		p.MagicToken1=CONTROL_MAGIC_TOKEN;
		p.MagicToken2=CONTROL_MAGIC_TOKEN;
		p.PacketType=CONTROL_PACKET_SET_BUFF;
		p.PacketID=Packets[i].PacketID;
		p.BuffID=((JobSetBuff*)job->JobVars)->BuffID;
		p.Offset=j*MAX_BUFFER_SEGMENT;
		if ((p.Offset+MAX_BUFFER_SEGMENT)>((JobSetBuff*)job->JobVars)->TargetBuffLen){
			p.Len=((JobSetBuff*)job->JobVars)->TargetBuffLen % MAX_BUFFER_SEGMENT;
		}else{
		 	p.Len=MAX_BUFFER_SEGMENT;
		}	
		p.Flags=0;
		memcpy(p.Data, ((JobSetBuff*)job->JobVars)->TargetBuff + p.Offset, p.Len); 


		((unsigned int*)Packets[i].RawPacket)[0]=CONTROL_MAGIC_TOKEN;
		job->Target->Crypto->Encrypt((struct crypto_record*)job->Target->Crypto, (char*)&p+4, Packets[i].RawPacket+4, sizeof(ControlPacketSetBuffer)-4);

		/*send it off*/
		hm_resend_packet(&Packets[i]);
	}	
}

/*****************************************************
* Start a command job running
*****************************************************/
int hm_job_command_start(JobRecord* job){
	int							i;
	ControlPacketCommand		p;
	int							count;
		
	/*find a free packet*/
	for (i=0;i<MAX_PACKETS;i++) if (!Packets[i].InUse) break;
	if (i==MAX_PACKETS) return -1;
	
	/*build a packet*/
	Packets[i].InUse=TRUE;
	bzero(Packets[i].RawPacket, MAX_PACKET_SIZE);
	Packets[i].PacketSize=sizeof(ControlPacketCommand);
	Packets[i].PacketID=random()*65535*65535;
	Packets[i].RetriesLeft=NUM_RETRIES+1;
	Packets[i].LastTime=0;
	Packets[i].Job=job;
	Packets[i].Callback=hm_command_packet;
	
	p.MagicToken1=CONTROL_MAGIC_TOKEN;
	p.MagicToken2=CONTROL_MAGIC_TOKEN;
	p.PacketType=CONTROL_PACKET_COMMAND;
	p.BuffID=((JobCommand*)job->JobVars)->BuffID;
	p.CommandID=((JobCommand*)job->JobVars)->CommandID;
	p.Flags=((JobCommand*)job->JobVars)->Flags;
	memcpy(p.Data,((JobCommand*)job->JobVars)->Data,230);
	p.PacketID=Packets[i].PacketID;

	((unsigned int*)Packets[i].RawPacket)[0]=CONTROL_MAGIC_TOKEN;
	job->Target->Crypto->Encrypt((struct crypto_record*)job->Target->Crypto, (char*)&p+4, Packets[i].RawPacket+4, sizeof(ControlPacketCommand)-4);

	/*send it off*/
	return hm_resend_packet(&Packets[i]);
}


/*****************************************************
* Start a file sync job running
*****************************************************/
int hm_job_file_sync_start(JobRecord* job){
	int							i,j;
	ControlPacketBuffRequest	p;
	int							count;
	int							num_packets;
	int							first_packet;
	struct stat					st;
	JobFileSync*				jf;

	/*get the file stat*/
	jf=(JobFileSync*)job->JobVars;
	if (stat(jf->FName, &st)==-1){
		/*create file*/
		printf("Creating new file\n");
		jf->fp=fopen(jf->FName, "w+");
	}else{
		if (st.st_size>=jf->BuffLen){
			printf("No update necessary\n");
			return TRUE;
		}
		/*append file*/
		printf("Syncing existing file\n");
		jf->fp=fopen(jf->FName, "r+");
	}
	
	/*mark the existing blocks as gotten*/
	for (i=0;i<(st.st_size/MAX_BUFFER_SEGMENT);i++){
		jf->BlockStatus[i]=BLOCK_STATUS_DONE;
	}
		
	/*send off a bunch of packets*/
	num_packets=((jf->BuffLen-st.st_size)/MAX_BUFFER_SEGMENT)+1;
	first_packet=st.st_size/MAX_BUFFER_SEGMENT;
	if (num_packets>MAX_PACKET_PER_JOB) num_packets=MAX_PACKET_PER_JOB;
		
#ifdef DEBUG		
	printf("Have to get %i packets\n",num_packets);	
#endif	
	/*find a free packet*/
	for (j=0;j<num_packets;j++){
#ifdef DEBUG	
		printf("Starting packet %i\n",j);
#endif		
		for (i=0;i<MAX_PACKETS;i++) if (!Packets[i].InUse) break;
		if (i==MAX_PACKETS) return -1;
		
		/*mark the packet as getting*/
		jf->BlockStatus[j+first_packet]=BLOCK_STATUS_GETTING;
		
		/*build a packet*/
		Packets[i].InUse=TRUE;
		bzero(Packets[i].RawPacket, MAX_PACKET_SIZE);
		Packets[i].PacketSize=sizeof(ControlPacketBuffRequest);
		Packets[i].PacketID=random()*65535*65535;
		Packets[i].RetriesLeft=NUM_RETRIES+1;
		Packets[i].LastTime=0;
		Packets[i].Job=job;
		Packets[i].Callback=hm_file_sync_packet;
		
		p.MagicToken1=CONTROL_MAGIC_TOKEN;
		p.MagicToken2=CONTROL_MAGIC_TOKEN;
		p.PacketType=CONTROL_PACKET_BUFF_REQUEST;
		p.PacketID=Packets[i].PacketID;
		p.BuffID=jf->BuffID;
		p.Offset=(j+first_packet)*MAX_BUFFER_SEGMENT;
		p.Len=MAX_BUFFER_SEGMENT;
		p.Flags=0;

		((unsigned int*)Packets[i].RawPacket)[0]=CONTROL_MAGIC_TOKEN;
		job->Target->Crypto->Encrypt((struct crypto_record*)job->Target->Crypto, (char*)&p+4, Packets[i].RawPacket+4, sizeof(ControlPacketBuffRequest)-4);
		
		/*send it off*/
		hm_resend_packet(&Packets[i]);
	}	
	
#ifdef DEBUG	
	printf("All Done\n");
#endif	
	return TRUE;
}



/*****************************************************
* Start up a new job
* returns jobid of the job created or -1 on error
*****************************************************/
int hm_exec_job(struct in_addr IP, char JobType, void* JobVars, JobCallback RFunc){
	int i;

	/*find a free job*/
	for (i=0;i<MAX_JOBS;i++) if (!Jobs[i].InUse) break;
	if (i==MAX_JOBS) return -1;
	
	/*set up the job*/
	Jobs[i].Target=hm_get_target_from_ip(IP);
	if (!Jobs[i].Target) return -1;
	Jobs[i].JobType=JobType;
	Jobs[i].JobVars=JobVars;
	Jobs[i].InUse=TRUE;
	Jobs[i].BlockingResult=BLOCKING_RESULT_WAITING;
	Jobs[i].Callback=RFunc;
	
	/*execute the job*/
	switch (Jobs[i].JobType){
	case JOB_TYPE_STAT:
		if (!hm_job_stat_start(&Jobs[i])) return -1;
		break;
	case JOB_TYPE_PING:
		if (!hm_job_ping_start(&Jobs[i])) return -1;
		break;	
	case JOB_TYPE_GET_BUFF:
		if (!hm_job_get_buff_start(&Jobs[i])) return -1;
		break;	
	case JOB_TYPE_SET_BUFF:
		if (!hm_job_set_buff_start(&Jobs[i])) return -1;
		break;		
	case JOB_TYPE_COMMAND:
		if (!hm_job_command_start(&Jobs[i])) return -1;	
		break;
	case JOB_TYPE_FILE_SYNC:
		if (!hm_job_file_sync_start(&Jobs[i])) return -1;		
		break;
	default:
		return -1;
	}
	
	return i;
}

/*****************************************************
* Non-Blocking Call to get a packet off fd
*****************************************************/
int hn_read_sock(){
	char						Buff[65536];
	long int					count, recvcount;	
	struct sockaddr_in			sender;
	int							size;
	TargetRecord*				target;
	int							i;
	ControlPacket*				cp;

	/*read in the packet*/
	size=sizeof(struct sockaddr);
	recvcount=recvfrom(GlobalFD, Buff, 65535, 0, (struct sockaddr*)&sender, &size);
	if (recvcount==-1) return BUFFER_FLAG_NOT_READY;

#ifdef DEBUGPATH
	printf("Got a Packet\n");
#endif

	/*decrypt the packet*/
	target=hm_get_target_from_ip(sender.sin_addr);
	if (!target){
		printf("I don't know %s\n", inet_ntoa(sender.sin_addr));
		return 0;
	}
	target->Crypto->Decrypt((struct crypto_record*)target->Crypto, Buff+4, Buff+4, recvcount-4);
	
	cp=(ControlPacket*)Buff;
	if (cp->MagicToken2!=CONTROL_MAGIC_TOKEN){
		printf("Magic token didn't match, corrupt packet?\n");
		return BUFFER_FLAG_KEY_ERROR;
	}
	
#ifdef DEBUGPATH
	printf("Packet was Decrypted\n");
#endif	

	/*find the correct handler*/
	for (i=0;i<MAX_PACKETS;i++){
		//if (Packets[i].InUse) printf("State %i %u vs. %u\n",i,Packets[i].PacketID, cp->PacketID);
		if (Packets[i].InUse && Packets[i].PacketID==cp->PacketID) break;
	}
	if (i==MAX_PACKETS){ 
		printf("Couldn't find a state for this packet %u\n", cp->PacketID);
		return BUFFER_FLAG_NO_STATE;
	}	
	
	/*call the handler*/
	Packets[i].InUse=FALSE;
	if (Packets[i].Callback) return Packets[i].Callback(&Packets[i], TRUE, Buff, recvcount, Packets[i].User);
}

/*****************************************************
* Non-Blocking Call to get the Buffer Status
*****************************************************/
int hn_get_stat(struct in_addr IP, unsigned int BuffID, ControlPacketStatReply* stat, JobCallback RFunc, void* User){
	JobStat*	js;
	
	js=(JobStat*)calloc(sizeof(JobStat),1);
	js->BuffID=BuffID;	
	js->Stat=stat;
	js->User=User;
	return hm_exec_job(IP, JOB_TYPE_STAT, (void*)js, RFunc);
}

/*****************************************************
* Blocking Call to get the Buffer Status
*****************************************************/
int hb_get_stat(struct in_addr IP, unsigned int BuffID, ControlPacketStatReply* stat){
	int JobID;
	
	JobID=hn_get_stat(IP, BuffID, stat, NULL, NULL);
	while (Jobs[JobID].BlockingResult==BLOCKING_RESULT_WAITING){
		hn_read_sock();
		if (hm_resend_all()) printf("Resending...\n");
		usleep(100);
	}
		
	Jobs[JobID].InUse=FALSE;
	free(Jobs[JobID].JobVars);
	return Jobs[JobID].BlockingResult;
}

/*****************************************************
* Non-Blocking Call to ping
*****************************************************/
int hn_ping(struct in_addr IP, char* Message, JobCallback RFunc, void* User){
	JobPing*	jp;
	
	jp=(JobPing*)calloc(sizeof(JobPing),1);
	snprintf(jp->Message, 244, Message);
	jp->User=User;
	return hm_exec_job(IP, JOB_TYPE_PING, (void*)jp, RFunc);
}

/*****************************************************
* Blocking Call to ping
*****************************************************/
int hb_ping(struct in_addr IP, char* Message){
	int JobID;
	
	JobID=hn_ping(IP, Message, NULL, NULL);
	if (JobID==-1) printf("Error Starting Job\n");
	while (Jobs[JobID].BlockingResult==BLOCKING_RESULT_WAITING){
		hn_read_sock();
		if (hm_resend_all()) printf("Resending...\n");
		usleep(100);
	}
	
	Jobs[JobID].InUse=FALSE;
	free(Jobs[JobID].JobVars);
	return Jobs[JobID].BlockingResult;
}

/*****************************************************
* Non-Blocking Call to get a buffer
*****************************************************/
int hn_get_buff(struct in_addr IP, unsigned int BuffID, char* TargetBuff, unsigned int TargetBuffLen, JobCallback RFunc, void* User){
	JobGetBuff*	jb;
	
#ifdef DEBUG
	printf("Starting Buff Job\n");
	printf("TargetBuff %p\n",TargetBuff);
#endif	
	
	jb=(JobGetBuff*)calloc(sizeof(JobGetBuff),1);
	jb->BuffID=BuffID;
	jb->TargetBuff=TargetBuff;
	jb->TargetBuffLen=TargetBuffLen;
	jb->BlockStatus=(char*)calloc((TargetBuffLen/MAX_BUFFER_SEGMENT)+1,1);
	jb->User=User;
	return hm_exec_job(IP, JOB_TYPE_GET_BUFF, (void*)jb, RFunc);
}

/*****************************************************
* Blocking Call to get a buffer
* TODO: FIX memory leak
*****************************************************/
int hb_get_buff(struct in_addr IP, unsigned int BuffID, char* TargetBuff, unsigned int TargetBuffLen){
	int JobID;
	
	JobID=hn_get_buff(IP, BuffID, TargetBuff, TargetBuffLen, NULL, NULL);
	if (JobID==-1) printf("Error Starting Job\n");
	while (Jobs[JobID].BlockingResult==BLOCKING_RESULT_WAITING){
		hn_read_sock();
		if (hm_resend_all()) printf("Resending...\n");
		usleep(100);
	}
	
	Jobs[JobID].InUse=FALSE;
	//free( ((JobGetBuff*)Jobs[JobID].JobVars)->BlockStatus);
	free(Jobs[JobID].JobVars);
	return Jobs[JobID].BlockingResult;
}

/*****************************************************
* Non-Blocking Call to set a buffer
*****************************************************/
int hn_set_buff(struct in_addr IP, unsigned int BuffID, char* TargetBuff, unsigned int TargetBuffLen, JobCallback RFunc, void* User){
	JobSetBuff*	jb;
	
	jb=(JobSetBuff*)calloc(sizeof(JobSetBuff),1);
	jb->BuffID=BuffID;
	jb->TargetBuff=TargetBuff;
	jb->TargetBuffLen=TargetBuffLen;
	jb->BlockStatus=(char*)calloc((TargetBuffLen/MAX_BUFFER_SEGMENT)+1,1);
	jb->User=User;
	return hm_exec_job(IP, JOB_TYPE_SET_BUFF, (void*)jb, RFunc);
}

/*****************************************************
* Blocking Call to set a buffer
* TODO: FIX memory leak
*****************************************************/
int hb_set_buff(struct in_addr IP, unsigned int BuffID, char* TargetBuff, unsigned int TargetBuffLen){
	int JobID;
	
	JobID=hn_set_buff(IP, BuffID, TargetBuff, TargetBuffLen, NULL, NULL);
	if (JobID==-1) printf("Error Starting Job\n");
	while (Jobs[JobID].BlockingResult==BLOCKING_RESULT_WAITING){
		hn_read_sock();
		if (hm_resend_all()) printf("Resending...\n");
		usleep(100);
	}
	
	Jobs[JobID].InUse=FALSE;
	//free( ((JobGetBuff*)Jobs[JobID].JobVars)->BlockStatus);
	free(Jobs[JobID].JobVars);
	return Jobs[JobID].BlockingResult;
}

/*****************************************************
* Non-Blocking Call to send a command
*****************************************************/
int hn_send_command(struct in_addr IP, unsigned int BuffID, unsigned int CommandID, int Flags, char* Data, int DataLen, ControlPacketCommandReply* reply, JobCallback RFunc, void* User){
	JobCommand*	jc;
	
	jc=(JobCommand*)calloc(sizeof(JobCommand),1);
	jc->BuffID=BuffID;	
	jc->CommandID=CommandID;
	jc->Flags=Flags;
	memcpy(jc->Data, Data, DataLen);
	jc->Reply=reply;
	jc->User=User;
	return hm_exec_job(IP, JOB_TYPE_COMMAND, (void*)jc, RFunc);
}

/*****************************************************
* Blocking Call to send a command
*****************************************************/
int hb_send_command(struct in_addr IP, unsigned int BuffID, unsigned int CommandID, int Flags, char* Data, int DataLen, ControlPacketCommandReply* reply){
	int JobID;
	
	JobID=hn_send_command(IP, BuffID, CommandID, Flags, Data, DataLen, reply,NULL, NULL);
	while (Jobs[JobID].BlockingResult==BLOCKING_RESULT_WAITING){
		hn_read_sock();
		if (hm_resend_all()) printf("Resending...\n");
		usleep(100);
	}
		
	Jobs[JobID].InUse=FALSE;
	free(Jobs[JobID].JobVars);
	return Jobs[JobID].BlockingResult;
}

/*****************************************************
* Non-Blocking Call to sync a file with a buffer
*****************************************************/
int hn_file_sync(struct in_addr IP, unsigned int BuffID, unsigned int BuffLen, char* FName, JobCallback RFunc, void* User){
	JobFileSync*	jf;
	
	jf=(JobFileSync*)calloc(sizeof(JobFileSync),1);
	jf->BuffID=BuffID;
	jf->BuffLen=BuffLen;
	snprintf(jf->FName, 1024, FName);
	jf->BlockStatus=(char*)calloc((BuffLen/MAX_BUFFER_SEGMENT)+1,1);
	jf->User=User;
	return hm_exec_job(IP, JOB_TYPE_FILE_SYNC, (void*)jf, RFunc);
}

/*****************************************************
* Blocking Call to sync a file with a buffer
*****************************************************/
int hb_file_sync(struct in_addr IP, unsigned int BuffID, unsigned int BuffLen, char* FName){
	int JobID;
	
	JobID=hn_file_sync(IP, BuffID, BuffLen, FName, NULL, NULL);
	if (JobID==-1) printf("Error Starting Job\n");
	while (Jobs[JobID].BlockingResult==BLOCKING_RESULT_WAITING){
		hn_read_sock();
		if (hm_resend_all()) printf("Resending...\n");
		usleep(100);
	}
	
	Jobs[JobID].InUse=FALSE;
	//free( ((JobGetBuff*)Jobs[JobID].JobVars)->BlockStatus);
	free(Jobs[JobID].JobVars);
	return Jobs[JobID].BlockingResult;
}

/*****************************************************
* Alias of the above
*****************************************************/
int hb_get_file(struct in_addr IP, unsigned int BuffID, unsigned int BuffLen, char* FName){
	return hb_file_sync(IP, BuffID, BuffLen, FName);
}


/*+++++++++++++++++++++++++++++Config File++++++++++++++++++*/

/**************************************************
* Read the next line from the file into Buff
**************************************************/
int hm_get_next_line(int fd, char* Buff, int BuffLen){
	int		count;
	char	c;
	int		num_read;
	char*	this;
	
	num_read=0;
	this=Buff;
	while ((count=read(fd, &c, 1))!=-1){
		if (count==0) break;
		*this=c;
		if (c!=0x20){
		 this++;
		 num_read++;
		} 
		if (c==0x0A) break;
		if (num_read==BuffLen) break;
	}
	this--;
	*this=0;
	
	return num_read;
	
}

/**************************************************
* Get the next field out of the string
**************************************************/
int hm_get_next_field(char* InBuff, char StartChar, char EndChar, char* OutBuff, int OutBuffLen){
	int		count;
	char	c;
	int		num_read;
	char	*start, *end;
	
	num_read=0;
	start=InBuff;

	bzero(OutBuff, OutBuffLen);

	/*seek to the start char*/
	if (StartChar!=0){
		while (*start!=0x0 && *start!=StartChar) start++;
		if (*start==0x0) return 0;
		start++;
	}
	
	/*seek to the ending char*/
	end=start+1;
	while (*end!=0x0 && *end!=EndChar){
		num_read++;
		if (num_read==OutBuffLen) break;
		end++;
	}	
	num_read++;
	memcpy(OutBuff, start, num_read);
	
	return num_read;	
}

/************************************************
* Read in the contents of the config file
************************************************/
int hm_read_config(char* config_file){
	int 	fd;
	char	LineBuff[1024];
	char	FieldBuff[1024];
	char	IntBuff[9];
	int		count;
	int		i;
	int		TargetNum;
	
	struct in_addr	IP;
	
	TargetNum=0;
	fd = open(config_file, O_RDONLY);
	if (fd==0){
		printf("Couldn't open config file\n");
		return FALSE;
	}
	
	while (hm_get_next_line(fd, LineBuff, 1024)){
		if (LineBuff[0]=='#') continue;
		if (LineBuff[0]==0x0A) continue;
		if (LineBuff[0]==0x0) continue;
		
		hm_get_next_field(LineBuff, 0,'(',FieldBuff, 1024);
		if (!inet_aton(FieldBuff, &IP)){
			printf("Bad IP address %s\n",FieldBuff);
			return FALSE;
		}
		
		hm_get_next_field(LineBuff, '(',',',FieldBuff, 1024);
		if (strncasecmp(FieldBuff,"None",4)==0){
			hm_get_next_field(LineBuff, ',',')',FieldBuff, 1024);
			if (AddKey(CRYPTO_TYPE_NONE, (int)IP.s_addr, FieldBuff)==-1){
				printf("Failed to add null key %s\n", FieldBuff);
				return FALSE;
			}
		}else if (strncasecmp(FieldBuff,"TwoFish",7)==0){
			hm_get_next_field(LineBuff, ',',')',FieldBuff, 1024);
			if (AddKey(CRYPTO_TYPE_TWOFISH, (int)IP.s_addr, FieldBuff)==-1){
				printf("Failed to add twofish key %s\n", FieldBuff);
				return FALSE;
			}
		}else if (strncasecmp(FieldBuff,"3DES",4)==0){
			hm_get_next_field(LineBuff, ',',')',FieldBuff, 1024);
			if (AddKey(CRYPTO_TYPE_3DES, (int)IP.s_addr, FieldBuff)==-1){
				printf("Failed to add 3DES key %s\n", FieldBuff);
				return FALSE;
			}
		}else{
			printf("I don't understand key type %s\n",FieldBuff);
			return FALSE;
		}
		
		hm_get_next_field(LineBuff, ')',0  ,Targets[TargetNum].Name, 512);
		Targets[TargetNum].InUse=TRUE;
		Targets[TargetNum].IP=IP;
		Targets[TargetNum].Crypto=GetCryptoFromID((int)IP.s_addr);
		
		
#ifdef KEEP_UNTIL_I_ERASE		
		if (strncasecmp(LineBuff, "target", 6)==0){
			/*read in a twofish key*/
			bzero(FieldBuff,1024);
			count=hm_get_next_field(LineBuff, '(',',',FieldBuff, 1024);	
			if (!inet_aton(FieldBuff, &Targets[TargetNum].IP)) continue;
			bzero(FieldBuff,1024);
			count=hm_get_next_field(LineBuff, ',',')',FieldBuff, 1024);	
			if (count!=((KeySize/32)*8)){
				printf("Wrong key size was %i characters expected %i characters\n",count,((KeySize/32)*8));
				continue;
			}	
			for (i=0;i<(KeySize/32);i++){
				memcpy(IntBuff, FieldBuff+(i*8),8);
				IntBuff[9]=0x0;
				Targets[TargetNum].Key.ki.key32[i]=strtol(IntBuff, NULL, 16); 
			}
			reKey(&Targets[TargetNum].Key.ki);
			bzero(FieldBuff,1024);
			count=hm_get_next_field(LineBuff, ')',0,FieldBuff, 1024);	
			snprintf(Targets[TargetNum].Name,512,"%s",FieldBuff);
			Targets[TargetNum].InUse=TRUE;
			TargetNum++;
		}else{
			printf("Unrecognized Keyword %s\n",LineBuff);
		}
#endif		
		TargetNum++;
	}
	
	close(fd);
	
	return TRUE;
}
