#include "control.h"
#include "hogwash.h"
#include "packet.h"
#include <netinet/in.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
/*for packet statistics from snort*/
#include "../snort-1.8.6/snort.h"

/*structs used to build return packets*/
typedef struct h_ethernet
{
    u_int8_t ether_dst[6];
    u_int8_t ether_src[6];
    u_int16_t ether_type;

}HEthernet;

typedef struct h_ip
{
    u_int8_t ip_hlen:4, ip_ver:4;
    u_int8_t ip_tos;        /* type of service */
    u_int16_t ip_len;       /* datagram length */
    u_int16_t ip_id;        /* identification  */
    u_int16_t ip_off;       /* fragment offset */
    u_int8_t ip_ttl;        /* time to live field */
    u_int8_t ip_proto;      /* datagram protocol */
    u_int16_t ip_csum;      /* checksum */
    struct in_addr ip_src;  /* source IP */
    struct in_addr ip_dst;  /* dest IP */
}HIP;

typedef struct h_tcp
{
    u_int16_t th_sport;     /* source port */
    u_int16_t th_dport;     /* destination port */
    u_int32_t th_seq;       /* sequence number */
    u_int32_t th_ack;       /* acknowledgement number */
    u_int8_t th_x2:4, th_off:4;
    u_int8_t th_flags;
    u_int16_t th_win;       /* window */
    u_int16_t th_sum;       /* checksum */
    u_int16_t th_urp;       /* urgent pointer */

}HTCP;

typedef struct h_udp
{
    u_int16_t uh_sport;
    u_int16_t uh_dport;
    u_int16_t uh_len;
    u_int16_t uh_chk;

}HUDP;

typedef struct h_icmp
{
    u_int8_t type;
    u_int8_t code;
    u_int16_t csum;
    u_int32_t data;
}HICMP;

#define	H_ICMP	1
#define	H_TCP	6
#define	H_UDP	17

#ifdef __USE_PTHREADS__
pthread_mutex_t	ControlMutex=PTHREAD_MUTEX_INITIALIZER;
#endif

extern PacketCount	pc;

/**************************************************
* Read the next line from the file into Buff
**************************************************/
int chogwash_get_next_line(int fd, char* Buff, int BuffLen){
	int		count;
	char	c;
	int		num_read;
	char*	this;
	
	num_read=0;
	this=Buff;
	while ((count=read(fd, &c, 1))!=-1){
		if (count==0) break;
		*this=c;
		if (c!=0x20){
		 this++;
		 num_read++;
		} 
		if (c==0x0A) break;
		if (num_read==BuffLen) break;
	}
	this--;
	*this=0;
	
	return num_read;
	
}

/**************************************************
* Get the next field out of the string
**************************************************/
int chogwash_get_next_field(char* InBuff, char StartChar, char EndChar, char* OutBuff, int OutBuffLen){
	int		num_read;
	char	*start, *end;
	
	num_read=0;
	start=InBuff;

	/*seek to the start char*/
	if (StartChar!=0){
		while (*start!=0x0 && *start!=StartChar) start++;
		if (*start==0x0) return 0;
		start++;
	}
	
	/*seek to the ending char*/
	end=start+1;
	while (*end!=0x0 && *end!=EndChar){
		num_read++;
		if (num_read==OutBuffLen) break;
		end++;
	}	
	num_read++;
	memcpy(OutBuff, start, num_read);
	
	return num_read;
	
}

/************************************************
* Read in the contents of the config file
************************************************/
int chogwash_read_config(char* config_file){
	int 	fd;
	char	LineBuff[1024];
	
	char	CryptoType;
	char*	args;
	

#ifdef DEBUG
	printf("Reading in control config file\n");
#endif
	
	fd = open(config_file, O_RDONLY);
	if (fd==0){
		printf("Couldn't open config file\n");
		return FALSE;
	}
	
	while (chogwash_get_next_line(fd, LineBuff, 1024)){
		if (LineBuff[0]=='#') continue;
		if (LineBuff[0]==0x0A) continue;
		if (LineBuff[0]==0x0) continue;
		
		if (strncasecmp(LineBuff, "None", 4)==0){
			/*NULL crypto*/
			CryptoType=CRYPTO_TYPE_NONE;
			args=LineBuff+4;
		}else if (strncasecmp(LineBuff, "TwoFish", 7)==0){
			/*twofish crypto*/
			CryptoType=CRYPTO_TYPE_TWOFISH;
			args=LineBuff+7;
		}else if (strncasecmp(LineBuff, "3DES", 4)==0){
			/*3des crypto*/
			CryptoType=CRYPTO_TYPE_3DES;
			args=LineBuff+4;
		}else{
			printf("Unrecognized crypto type %s\n", LineBuff);
			return FALSE;
		}
		
		/*seek to the next non-whitespace*/
		while (*args==' ' || *args=='\t') args++;
		
		return AddKey(CryptoType, 0, args);

	}
	close(fd);
	
	return TRUE;
}

/****************************************************
* Create a new file buffer
****************************************************/
int create_file_buffer(char* fname, unsigned int BuffID){
	int i;
	char*	startver;
	
	for (i=0;i<MAX_BUFFERS;i++){
		if (!Buffers[i].InUse) break;
	}
	
	if (i==MAX_BUFFERS) return FALSE;
	if (strlen(fname)==0) return FALSE;
	
	printf("Adding into Buffer %i\n",i);
	
	/*see if the buffer has a version*/
	startver=strstr(fname, ".v");
	if (!startver){
		strncpy(Buffers[i].FName, fname, 1024);
		Buffers[i].Version=-1;
	}else{
		memcpy(Buffers[i].FName, fname, startver-fname);
		Buffers[i].Version=atoi(startver+2);
	}
	
	Buffers[i].BuffID=BuffID;
	Buffers[i].BuffType = BUFFER_TYPE_SHARED_FILE;
	Buffers[i].DestroyTime=0;
	Buffers[i].InUse=TRUE;
	
	printf("FileName is %s\n",Buffers[i].FName);
	printf("Verion is %i\n",Buffers[i].Version);
	
	return TRUE;
}

/***************************************************
* Parse the command line options for the control channel
***************************************************/
void control_cmd_line(int argc, char**argv){
	int i;
	
	for (i=0;i<argc;i++){
		if (strncasecmp("-k", argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=='-')){
				printf("Syntax -k <control config file>\n");
				exit(1);
			}
			CV.UseControl=TRUE;
			snprintf(CV.ConfigFileName, 1024, argv[i+1]);
		}
	}
}

/****************************************************
* Set up everything to do control channel stuff
****************************************************/
void control_init(int argc, char* argv[]){
	int j;
	
#ifdef DEBUG	
	printf("Initializing Control Channel\n");
	printf("Parsing Control args\n");
#endif
	
	bzero(&CV, sizeof(struct control_vars));
	control_cmd_line(argc, argv);
	
	/*create stock buffers*/
	bzero(Buffers, sizeof(XBuff) * MAX_BUFFERS);
	for (j=0;j<MAX_BUFFERS;j++){
		Buffers[j].Version=-1;
	}
	create_file_buffer(ConfigFile, STOCK_BUFFER_RULES);
	create_file_buffer(LastAlertFileName, STOCK_BUFFER_ALERT);
	create_file_buffer(LastLogFileName, STOCK_BUFFER_LOG);
	create_file_buffer(ConfigFile, STOCK_BUFFER_NEW_RULES);
	create_file_buffer(LastUnifiedAlertFileName, STOCK_BUFFER_UALERT);
	create_file_buffer(LastUnifiedLogFileName, STOCK_BUFFER_ULOG);
	create_file_buffer(SidPath, STOCK_BUFFER_SID);

	CryptoInit();	
	
	chogwash_read_config(CV.ConfigFileName);
}

/***********************************************
* Sends a packet back to the sender
* TODO: Finish building packet
************************************************/
int BuildReturnPacket(int to_interface, char* OldPacket, int OldLen, char* Message, int MessageLen, int Span){
	unsigned char 	Transport;
	char			NewBuff[PACKET_HOLDER_SIZE];
	int				PacketLen;
	
	HEthernet	*hethn, *hetho;
	HIP			*hipn, *hipo;
	HTCP		*htcpn, *htcpo;
	HUDP		*hudpn, *hudpo;
	HICMP		*hicmpn, *hicmpo;

	bzero(NewBuff,PACKET_HOLDER_SIZE);

	/*determine transport layer*/
	Transport=*(unsigned char*)(OldPacket+23);
	
	/*cast struct over everything*/
	hetho=(HEthernet*)OldPacket;
	hethn=(HEthernet*)NewBuff;
	hipo=(HIP*)(OldPacket+sizeof(HEthernet));
	hipn=(HIP*)(NewBuff+sizeof(HEthernet));
	switch (Transport){
	case H_ICMP:
		hicmpo=(HICMP*)(OldPacket+sizeof(HEthernet)+sizeof(HIP));
		hicmpn=(HICMP*)(NewBuff+sizeof(HEthernet)+sizeof(HIP));
		break;
	case H_TCP:
		htcpo=(HTCP*)(OldPacket+sizeof(HEthernet)+sizeof(HIP));
		htcpn=(HTCP*)(NewBuff+sizeof(HEthernet)+sizeof(HIP));
	case H_UDP:
		hudpo=(HUDP*)(OldPacket+sizeof(HEthernet)+sizeof(HIP));
		hudpn=(HUDP*)(NewBuff+sizeof(HEthernet)+sizeof(HIP));
	default:
		return FALSE;
	}

	/*build the ethernet header*/
	memcpy(hethn->ether_dst, hetho->ether_src, 6);
	memcpy(hethn->ether_src, hetho->ether_dst, 6);
	hethn->ether_type=hetho->ether_type;
	/*done building ethernet header*/
	
	/*build the ip header*/
	hipn->ip_ver=4;
	hipn->ip_hlen=5;
	hipn->ip_tos=0;
	hipn->ip_len=htons(sizeof(HIP)+MessageLen);
	hipn->ip_id=555*rand();
	hipn->ip_off=0;
	hipn->ip_ttl=255;
	hipn->ip_proto=Transport;
	hipn->ip_csum=0;
	hipn->ip_src=hipo->ip_dst;
	hipn->ip_dst=hipo->ip_src;
	/*done building ip header*/
	
	/*build the transport header*/
	switch (Transport){
	case H_ICMP:
		PacketLen=sizeof(HEthernet)+sizeof(HIP)+sizeof(HICMP)+MessageLen;
		break;
	case H_TCP:
		htcpn->th_sport=htcpo->th_dport;
		htcpn->th_dport=htcpo->th_sport;
		htcpn->th_seq=htcpo->th_ack;
		htcpn->th_ack=htcpo->th_seq+1;
		htcpn->th_x2=5;
		htcpn->th_off=0;
		htcpn->th_flags=0;
		htcpn->th_win=128;
		htcpn->th_sum=0;
		htcpn->th_urp=0;
		PacketLen=sizeof(HEthernet)+sizeof(HIP)+sizeof(HTCP)+MessageLen;
		break;
	case H_UDP:
		hudpn->uh_sport=hudpo->uh_dport;
		hudpn->uh_dport=hudpo->uh_sport;
		hudpn->uh_len=htons(sizeof(HUDP)+MessageLen);
		hudpn->uh_chk=0;
		PacketLen=sizeof(HEthernet)+sizeof(HIP)+sizeof(HUDP)+MessageLen;
		break;
	}
	/*done building the transport header*/
	
	/*calculate the checksums*/
	
	/*done calculating the checksums*/
	
	write_packet(to_interface, NewBuff, PacketLen, Span);
	
	return TRUE;
}

/*******************************************
* Read the contents of a file into a buffer
********************************************/
int ReadFile(char* fname, char* Buff, unsigned int offset, unsigned int len, int* flags){
	FILE*			fd;
	int				count;
	struct stat		st;
	unsigned int	read_len;
	
	printf("@@@@@@Requested %u bytes starting at %u\n",len, offset);
	
	*flags=BUFFER_FLAG_OK;
	if (stat(fname, &st)!=0){
		*flags=BUFFER_FLAG_NOT_FOUND;
		printf("BUFFER_FLAG_NOT_FOUND\n");
		printf("Error Number is %i\n",errno);
		printf("fname is %s\n",fname);
		switch (errno){
		case EBADF:
			printf("EBADF\n");
			break;
		case ENOENT:
			printf("ENOENT\n");
			break;	
		default:
			printf("No Idea\n");
		}
		return 0;
	}
	if (offset>st.st_size){
		*flags=BUFFER_FLAG_END_OF_BUFFER;
		printf("st_size %i\n",(int)st.st_size);
		printf("offset  %i\n",(int)offset);
		printf("BUFFER_FLAG_END_OF_BUFFER\n");
		return 0;
	}
	fd=fopen(fname, "r");
	if (!fd){
		*flags=BUFFER_FLAG_OPEN_ERROR;
		printf("BUFFER_FLAG_OPEN_ERROR\n");	
		return 0;
	}
	if (fseek(fd, offset, SEEK_SET)!=0){
		*flags=BUFFER_FLAG_SEEK_ERROR;
		printf("BUFFER_FLAG_SEEK_ERROR\n");
		return 0;
	} 
	read_len=len;
	if (offset+len > st.st_size){
		read_len=st.st_size-offset;
		printf("st_size %i\n",(int)st.st_size);
		printf("offset  %i\n",(int)offset);
		printf("changed length to %i\n",read_len);
	}
	count=fread(Buff, read_len, 1, fd);
	fclose(fd);
	
	printf("I read %s\n",Buff);
	printf("@@@@@@@Returning length %u\n",read_len);
	return read_len;
}

/*******************************************
* Write the contents of a buffer to a file
********************************************/
int WriteFile(char* fname, char* Buff, unsigned int offset, unsigned int len, int* flags){
	FILE*			fd;
	int				count;
	struct stat		st;
	
	printf("@@@@@@Writing file %s %u bytes starting at %u\n",fname, len, offset);
	
	*flags=BUFFER_FLAG_OK;
	if (stat(fname, &st)==-1){
		fd=fopen(fname, "w+");
	}else{
		fd=fopen(fname, "r+");
	}	
	if (!fd){
		*flags=BUFFER_FLAG_OPEN_ERROR;
		printf("BUFFER_FLAG_OPEN_ERROR\n");	
		return 0;
	}
	if (fseek(fd, offset, SEEK_SET)!=0){
		*flags=BUFFER_FLAG_SEEK_ERROR;
		printf("BUFFER_FLAG_SEEK_ERROR\n");
		return 0;
	} 
	count=fwrite(Buff, (int)len, 1, fd);
	fclose(fd);
	
	printf("I wrote %s\n",Buff);
	printf("@@@@@@@Returning length %u\n",len);
	return len;
}


/*****************************************
* Handle the ping request
******************************************/
int ProcessPing(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum, int Span){
	ControlPacketPing*		cpp;
	ControlPacketPingReply	cppr;
	int						to_interface;

	if (PayLen!=sizeof(ControlPacketPing)) return FALSE;
	
	cpp=(ControlPacketPing*)Pay;
	
#ifdef DEBUG1
	printf("Ping Packet:\n");
	printf("MagicToken1: %u\n",cpp->MagicToken1);
	printf("MagicToken2: %u\n",cpp->MagicToken2);
	printf("PacketType : %u\n",cpp->PacketType);
	printf("PacketID   : %u\n",cpp->PacketID);
	printf("Message    : %s\n",cpp->PingMessage);
#endif	
	
	cppr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cppr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cppr.PacketType=CONTROL_PACKET_PING_REPLY;
	cppr.PacketID=cpp->PacketID;
	sprintf(cppr.PingMessage, cpp->PingMessage);
	
	Cryptos[KeyNum].Encrypt((struct crypto_record*)&Cryptos[KeyNum], (char*)&cppr+4, (char*)&cppr+4, sizeof(ControlPacketPingReply)-4);
		
	if (ptp->from_interface==INTERNAL) to_interface=EXTERNAL; else to_interface=INTERNAL;
	return BuildReturnPacket(
		to_interface,
		ptp->raw_packet, 
		ptp->packet_len, 
		(char*)&cppr, 
		sizeof(ControlPacketPingReply),
		Span
	);
}


/*****************************************
* Handle the Buffer request
******************************************/
int ProcessBuff(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum, int Span){
	ControlPacketBuffRequest*		cpb;
	ControlPacketBuffReply			cpbr;
	int								i;
	char							RealFileName[1024];
	int								to_interface;

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketBuffRequest));
#endif
	if (PayLen!=sizeof(ControlPacketBuffRequest)) return FALSE;
	
	cpb=(ControlPacketBuffRequest*)Pay;
	
#ifdef DEBUG1
	printf("Buff Request Packet:\n");
	printf("MagicToken1: %u\n",cpb->MagicToken1);
	printf("MagicToken2: %u\n",cpb->MagicToken2);
	printf("PacketType : %u\n",cpb->PacketType);
	printf("PacketID   : %u\n",cpb->PacketID);
#endif	
	
	cpbr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpbr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpbr.PacketType=CONTROL_PACKET_BUFF_REPLY;
	cpbr.PacketID=cpb->PacketID;
	cpbr.BuffID=cpb->BuffID;
	cpbr.Offset=cpb->Offset;
	cpbr.Len=cpb->Len;
	cpbr.Flags=0;
	
	/*find the buffer*/
	for (i=0;i<MAX_BUFFERS; i++){
		if (Buffers[i].BuffID==cpbr.BuffID) break;
	}
	
	if (Buffers[i].Version==-1){
		sprintf(RealFileName, Buffers[i].FName);
	}else{
		sprintf(RealFileName, "%s.v%05i", Buffers[i].FName, Buffers[i].Version);
	}
	printf("Real file name is %s\n",RealFileName);
	
	if (i==MAX_BUFFERS){printf("Buffer ID not found\n");return 0;}
	if (cpbr.Len<=MAX_BUFFER_SEGMENT){
		cpbr.Len=ReadFile(RealFileName, cpbr.Data, cpbr.Offset, cpbr.Len, &cpbr.Flags);
	}

	Cryptos[KeyNum].Encrypt((struct crypto_record*)&Cryptos[KeyNum], (char*)&cpbr+4, (char*)&cpbr+4, sizeof(ControlPacketBuffReply)-4);

	if (ptp->from_interface==INTERNAL) to_interface=EXTERNAL; else to_interface=INTERNAL;	
	return BuildReturnPacket(
		to_interface,
		ptp->raw_packet, 
		ptp->packet_len, 
		(char*)&cpbr, 
		sizeof(ControlPacketBuffReply),
		Span
	);
}


/*****************************************
* Handle the Buffer Status request
******************************************/
int ProcessStat(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum, int Span){
	ControlPacketStatRequest*		cps;
	ControlPacketStatReply			cpsr;
	struct stat						st;
	int								i;
	char							RealFileName[1024];
	int								to_interface;

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketStatRequest));
#endif
	if (PayLen!=sizeof(ControlPacketStatRequest)) return FALSE;
	
	cps=(ControlPacketStatRequest*)Pay;
	
#ifdef DEBUG1
	printf("Stat Request Packet:\n");
	printf("MagicToken1: %u\n",cps->MagicToken1);
	printf("MagicToken2: %u\n",cps->MagicToken2);
	printf("PacketType : %u\n",cps->PacketType);
	printf("PacketID   : %u\n",cps->PacketID);
	printf("BufferID   : %u\n",cps->BuffID);
#endif	
	
	cpsr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpsr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpsr.PacketType=CONTROL_PACKET_STAT_REPLY;
	cpsr.PacketID=cps->PacketID;
	cpsr.BuffID=cps->BuffID;
	
	if (cps->BuffID!=0){
		/*find the buffer*/
		for (i=0;i<MAX_BUFFERS; i++){
			if (Buffers[i].InUse)
			if (Buffers[i].BuffID==cpsr.BuffID) break;
		}
	
		if (i==MAX_BUFFERS){printf("Couldn't find buffer\n");return 0;}
		
		if (Buffers[i].Version==-1){
			snprintf(RealFileName, 1024, Buffers[i].FName);
		}else{
			snprintf(RealFileName, 1024, "%s.v%05i",Buffers[i].FName, Buffers[i].Version);
		}
		if (stat(RealFileName, &st)!=0){
			printf("Couldn't stat file%i (Buff %u) *%s*\n",i,Buffers[i].BuffID, RealFileName);
		}
	
		/*Buffer stats*/
		cpsr.BuffType=BUFFER_TYPE_SHARED_FILE;
		cpsr.CreateTime=st.st_ctime;
		cpsr.ModifyTime=st.st_mtime;
		cpsr.TimeLeft=0;
		cpsr.Size=st.st_size;
		cpsr.BufferVersion=Buffers[i].Version;
	}
	
	/*snort stats*/
	cpsr.TCPCount=pc.tcp;
	cpsr.UDPCount=pc.udp;
	cpsr.ICMPCount=pc.icmp;
	cpsr.ARPCount=pc.arp;
	cpsr.IPV6Count=pc.ipv6;
	cpsr.IPXCount=pc.ipx;
	cpsr.OtherCount=pc.other;
	cpsr.AlertCount=pc.alert_pkts;
	cpsr.LogCount=pc.log_pkts;
	cpsr.DropCount=0;	/*Not yet implemented*/
		
	Cryptos[KeyNum].Encrypt((struct crypto_record*)&Cryptos[KeyNum], (char*)&cpsr+4, (char*)&cpsr+4, sizeof(ControlPacketStatReply)-4);	
	
	if (ptp->from_interface==INTERNAL) to_interface=EXTERNAL; else to_interface=INTERNAL;	
	return BuildReturnPacket(
		to_interface,
		ptp->raw_packet, 
		ptp->packet_len, 
		(char*)&cpsr, 
		sizeof(ControlPacketStatReply),
		Span
	);
}

/*****************************************
* Handle the Buffer Status request
******************************************/
int ProcessSetBuff(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum, int Span){
	ControlPacketSetBuffer*			cpsb;
	ControlPacketSetBufferReply		cpsbr;
	int								count;
	int								i;
	char							RealFileName[1024];
	int								to_interface;

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketSetBuffer));
#endif
	if (PayLen!=sizeof(ControlPacketSetBuffer)) return FALSE;
	
	cpsb=(ControlPacketSetBuffer*)Pay;
	
#ifdef DEBUG1
	printf("Set Buffer Packet:\n");
	printf("MagicToken1: %u\n",cpsb->MagicToken1);
	printf("MagicToken2: %u\n",cpsb->MagicToken2);
	printf("PacketType : %u\n",cpsb->PacketType);
	printf("PacketID   : %u\n",cpsb->PacketID);
	printf("BufferID   : %u\n",cpsb->BuffID);
#endif	

	/*find the buffer*/
	for (i=0;i<MAX_BUFFERS; i++){
		if (Buffers[i].InUse)
		if (Buffers[i].BuffID==cpsb->BuffID) break;
	}
	if (i==MAX_BUFFERS){printf("Couldn't find buffer\n");return 0;}
	
	if (Buffers[i].Version==-1){
		sprintf(RealFileName, Buffers[i].FName);
	}else{
		sprintf(RealFileName, "%s.v%05i", Buffers[i].FName, Buffers[i].Version);
	}
	printf("Real file name is %s\n",RealFileName);

	count=WriteFile(RealFileName, cpsb->Data, cpsb->Offset, cpsb->Len, &cpsb->Flags);
	
	cpsbr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpsbr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpsbr.PacketType=CONTROL_PACKET_SET_REPLY;
	cpsbr.Flags=cpsb->Flags;
	cpsbr.PacketID=cpsb->PacketID;
	cpsbr.Offset=cpsb->Offset;
	cpsbr.Len=cpsb->Len;

	Cryptos[KeyNum].Encrypt((struct crypto_record*)&Cryptos[KeyNum], (char*)&cpsbr+4, (char*)&cpsbr+4, sizeof(ControlPacketSetBufferReply)-4);	
		
	if (ptp->from_interface==INTERNAL) to_interface=EXTERNAL; else to_interface=INTERNAL;		
	return BuildReturnPacket(
		to_interface,
		ptp->raw_packet, 
		ptp->packet_len, 
		(char*)&cpsbr, 
		sizeof(ControlPacketSetBufferReply),
		Span
	);
		
	return FALSE;
}

/*****************************************
* Handle the Command request
******************************************/
int ProcessCommand(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum, int Span){
	ControlPacketCommand*			cpc;
	ControlPacketCommandReply		cpcr;
	int								i;
	int								to_interface;

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketCommand));
#endif
	if (PayLen!=sizeof(ControlPacketCommand)) return FALSE;
	
	cpc=(ControlPacketCommand*)Pay;
	
#ifdef DEBUG1
	printf("Command Packet:\n");
	printf("MagicToken1: %u\n",cpc->MagicToken1);
	printf("MagicToken2: %u\n",cpc->MagicToken2);
	printf("PacketType : %u\n",cpc->PacketType);
	printf("PacketID   : %u\n",cpc->PacketID);
	printf("BufferID   : %u\n",cpc->BuffID);
#endif	

	switch(cpc->CommandID){
	case COMMAND_RESTART:
		printf("Executing Restart Command\n");
		restart();
		return FALSE;
	case COMMAND_SET_BUFFER_VERSION:
		printf("Set Buffer Version\n");
		/*find the buffer*/
		for (i=0;i<MAX_BUFFERS; i++){
			if (Buffers[i].InUse)
			if (Buffers[i].BuffID==cpc->BuffID) break;
		}
		if (i==MAX_BUFFERS){printf("Couldn't find buffer\n");cpcr.Flags=FALSE;break;}

		/*set the new version*/
		Buffers[i].Version=*((int*)cpc->Data);
		printf("Version set to %i\n",Buffers[i].Version);
		cpcr.Flags=TRUE;
		break;
	default:
		printf("Unknown Command\n");
		return FALSE;
	}

	cpcr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpcr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpcr.PacketType=CONTROL_PACKET_COMMAND_REPLY;
	cpcr.PacketID=cpc->PacketID;
	cpcr.BuffID=cpc->BuffID;
	cpcr.CommandID=cpc->CommandID;

	Cryptos[KeyNum].Encrypt((struct crypto_record*)&Cryptos[KeyNum], (char*)&cpcr+4, (char*)&cpcr+4, sizeof(ControlPacketCommandReply)-4);	
		
	if (ptp->from_interface==INTERNAL) to_interface=EXTERNAL; else to_interface=INTERNAL;		
	return BuildReturnPacket(
		to_interface,
		ptp->raw_packet, 
		ptp->packet_len, 
		(char*)&cpcr, 
		sizeof(ControlPacketCommandReply),
		Span
	);
		
	return FALSE;
}


/****************************************/
/*find the control payload in the packet*/
/****************************************/
u_char* control_find_payload(u_char* pkt, unsigned int len, unsigned int* payload_len){
	u_char*			payload;
	unsigned short*	v1;	
	unsigned char*	v2;
	int				iplen,tcplen;
	
	if (len<256){
#ifdef DEBUG	
		printf("Packet was too short\n");
#endif		
		return NULL;
	}
	
	/*assume an ethernet packet right now*/
	v1=(unsigned short*)(pkt+12);
	if (*v1!=8){
#ifdef DEBUG	
		printf("Error in Ethernet Header. Not and IP packet\n");
#endif		
		return NULL;
	}
	
	v2=(unsigned char*)(pkt+23);
	iplen=((*(pkt+14)) & 0x0F)*4;
	switch (*v2){
	case 17:
#ifdef DEBUG
		printf("Was a UDP packet\n");
#endif
		payload=pkt+14+iplen+8;
		*payload_len=len-(14+iplen+8);
		break;	
	case 6:
#ifdef DEBUG
		printf("Was a TCP packet\n");
#endif
		tcplen=((*(pkt+14+iplen+13)) & 0xF0)/8 * 4;
		payload=pkt+14+iplen+tcplen;
		*payload_len=len-(14+iplen+tcplen);
		break;		
	case 1:
#ifdef DEBUG
		printf("Was an ICMP packet\n");
#endif
		payload=pkt+14+iplen+4;
		*payload_len=len-(14+iplen+4);
		break;		
	default:
#ifdef DEBUG
		printf("Can't tunnel over this packe type");
#endif
		return NULL;	
	}
	
	if (*((unsigned int*)payload)!=CONTROL_MAGIC_TOKEN){
#ifdef DEBUG
		printf("Didn't have the magic token\n");
#endif	
		return NULL;
	}
		
	printf("**This is a control packet**\n");
	return payload;
}


/*********************************************************
* All packets hit here
*********************************************************/
int control_func(PacketToProcess* ptp, int Span){
	u_char*				pay;
	unsigned int		payload_len;
	ControlPacket*		cp;
	int					i;
	char				Buff[65537];

#ifdef DEBUG	
	printf("Control got a packet\n");
#endif

#ifdef __USE_PTHREADS__
	pthread_mutex_lock(&ControlMutex);
#endif	
	
	/*check for the magic token*/
	pay=control_find_payload(ptp->raw_packet, ptp->packet_len, &payload_len);
	if (!pay){
#ifdef DEBUG	
		printf("Wasn't a control packet\n");
#endif		
#ifdef __USE_PTHREADS__
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_SENDON;
	}
	
#ifdef DEBUG
	printf("About to Decrypt the packet\n");
#endif	

	/*decrypt the packet*/
	memcpy(Buff, pay, payload_len);
	cp=(ControlPacket*)pay;	
	for (i=0;i<MAX_CRYPTO_RECORDS;i++){
		if (Cryptos[i].InUse)
		if (Cryptos[i].Decrypt((struct crypto_record*)&Cryptos[i], pay+4, pay+4, payload_len-4))
		if (cp->MagicToken2==CONTROL_MAGIC_TOKEN) break;
		/*put the buffer back for the next round*/
		memcpy(pay, Buff, payload_len);
	}		
	if (i==MAX_CRYPTO_RECORDS){
#ifdef __USE_PTHREADS__	
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_SENDON;	
	}	
	
#ifdef DEBUG1
	printf("I'm using key number %i\n",i);
#endif
			
#ifdef DEBUG
	printf("Handling packet type %u\n",cp->PacketType);
#endif
	
	/*handle the packet*/
	switch (cp->PacketType){
	case CONTROL_PACKET_PING:
		if (ProcessPing(ptp, pay, payload_len, i, Span)){
#ifdef __USE_PTHREADS__		
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_DROP;
		}else{
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_SENDON;
		}	
	case CONTROL_PACKET_PING_REPLY:
#ifdef __USE_PTHREADS__		
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_DROP;	
	case CONTROL_PACKET_BUFF_REQUEST:
		if (ProcessBuff(ptp, pay, payload_len, i, Span)){
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_DROP;
		}else{
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_SENDON;
		}	
	case CONTROL_PACKET_BUFF_REPLY:
#ifdef __USE_PTHREADS__		
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_DROP;	
	case CONTROL_PACKET_STAT_REQUEST:
		if (ProcessStat(ptp, pay, payload_len, i, Span)){
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_DROP;
		}else{
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_SENDON;
		}
	case CONTROL_PACKET_STAT_REPLY:
#ifdef __USE_PTHREADS__		
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_DROP;
	case CONTROL_PACKET_SET_BUFF:
		if (ProcessSetBuff(ptp, pay, payload_len, i, Span)){
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_DROP;
		}else{
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_SENDON;
		}	
	case CONTROL_PACKET_SET_REPLY:
#ifdef __USE_PTHREADS__		
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_DROP;	
	case CONTROL_PACKET_COMMAND:
		if (ProcessCommand(ptp, pay, payload_len, i, Span)){
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_DROP;
		}else{
#ifdef __USE_PTHREADS__			
			pthread_mutex_unlock(&ControlMutex);
#endif			
			return RESPONSE_SENDON;
		}	
	case CONTROL_PACKET_COMMAND_REPLY:
#ifdef __USE_PTHREADS__		
		pthread_mutex_unlock(&ControlMutex);
#endif		
		return RESPONSE_DROP;	
	
	default:
#ifdef DEBUG
		printf("I don't know how to handle PacketType %u\n",cp->PacketType);
#endif
#ifdef __USE_PTHREADS__
		pthread_mutex_unlock(&ControlMutex);
#endif
		return RESPONSE_SENDON;
	}
	
#ifdef __USE_PTHREADS__
	pthread_mutex_unlock(&ControlMutex);
#endif
	return RESPONSE_SENDON;	
}


